from __future__ import print_function
from mpi4py import MPI
from pygrisb.iface.wanniertb import w90
import numpy as np
import h5py, glob, scipy, warnings
from pygrisb.symm.unitary import get_u_csh2wan_all
from scipy.linalg import block_diag
from scipy.io import FortranFile
from pygrisb.estructure.gwannier import mpiget_bndev, \
        get_wannier_den_matrix_risb
from pygrisb.estructure.fermi import get_fermi_weight, get_fermi_level
from pygrisb.iface.wannierio import mpiget_wannier_data


def luse_h5(fname):
    try:
        f = h5py.File(fname, "r")
        f.close()
        use_h5 = True
    except:
        use_h5 = False
    return use_h5


def if_gwannier(corbs_list, delta_charge=0., wpath="../wannier",
        lpath="../lattice", wprefix="wannier", lprefix="mdl",
        lrot_list=None, iso=1, ispin=1, ismear=0, delta=0.0258,
        icycle=0):
    cell, _, kpts, include_bands, wfwannier_list, bnd_es = \
            mpiget_wannier_data(path=wpath)
    # total number of valence electrons
    n_elec = get_total_valence_elec("{}/{}_1.out".format(lpath, lprefix))
    n_elec -= max(0, (include_bands[0]-1)*(3-iso))
    symbols, atomic_positions = wget_symbols_positions(path=wpath,
            wprefix=wprefix)
    # convert to scaled position
    atomic_positions = np.asarray(atomic_positions).dot(
            np.linalg.inv(cell))
    numk = kpts.shape[0]
    # GMPI_x.h5 file
    kvec = set_kvec_para(numk)
    wk = 1./numk
    nbmax = wfwannier_list.shape[3]
    # spin degeneracy
    spin_deg = 3-max(ispin, iso)

    comm = MPI.COMM_WORLD
    myrank = comm.Get_rank()
    # wannier basis to complex spherical basis tranformation.
    u_wan2csh = get_wann2csh(nbmax, corbs_list)
    h1e_all = []
    nelectron = 0.
    # input file of dft bare band structure information for cygutz
    with h5py.File('BAREHAM_{}.h5'.format(myrank), 'w') as f:
        for isp in range(wfwannier_list.shape[0]):
            h1e_all.append(np.zeros((nbmax, nbmax), dtype=np.complex))
            for ik in range(kvec[0][2],kvec[0][2]+kvec[0][1]):
                # rescontruct dft hamiltonian matrix
                # in the wannier basis.
                # since it involves a downfolding,
                # some information outside of the frozen energy window
                # will be lost.
                hmat = wfwannier_list[isp][ik].T.conj().dot( \
                        np.diag(bnd_es[isp][ik])).dot( \
                        wfwannier_list[isp][ik])
                # from wannier basis to correlated orbital-ordered
                # complex spherical harmonics basis,
                # which is the convention used in the cygutz
                # initialization script.
                hmat = u_wan2csh.T.conj().dot(hmat).dot(u_wan2csh)
                # record the onsite one-body part
                h1e_all[isp] += hmat*wk
                # save the downfolded dft hamiltonian
                f['/IKP_{}/ISYM_1/HK0_SPIN1'.format(ik+1)] = hmat.T
                # get the eigen-value and eigen0vectors
                evals, evecs = np.linalg.eigh(hmat)
                # another way to evaluate total valence electrons
                # according to sangkook.
                nelectron += np.count_nonzero(evals < 0.)*(wk*spin_deg)
                # yes, here it is redundant here
                # but for the sake of consistent with wien2k interface.
                # here it implies the downfolding procedure is not necessary.
                f['/IKP_{}/ek0_spin1'.format(ik+1)] = evals
                f['/IKP_{}/T_PSIK0_TO_HK0_BASIS_SPIN{}'.format( \
                        ik+1, isp+1)] = evecs.T
    nelectron = comm.reduce(nelectron, op=MPI.SUM)
    h1e_all = np.asarray(h1e_all)
    h1e_all = comm.reduce(h1e_all, op=MPI.SUM)
    if myrank == 0:
        h1e_list = []
        for isp in range(wfwannier_list.shape[0]):
            h1e_list.append([])
            base = 0
            for corbs in corbs_list:
                norbs = len(corbs)
                h1e_list[isp].append(h1e_all[isp][base:base+norbs, \
                        base:base+norbs])
                base += norbs
        nelectron = int(nelectron+0.5)
        if np.abs(nelectron - n_elec) > 1.e-6:
            warnings.warn(" wannier valence electrons: {} vs {}!".format( \
                    nelectron, n_elec))
            n_elec = nelectron
        if icycle <= 1:
            wrt_ginit(symbols, cell, atomic_positions, u_wan2csh,
                    lrot_list=lrot_list)
        else:
            update_ginit(u_wan2csh)

        ne_list = [[nbmax, 1, nbmax] for k in range(numk)]
        wk_list = [wk for k in range(numk)]
        nelectron = n_elec + delta_charge
        wrt_gparambands(numk, nbmax, ne_list, wk_list, kpts, nelectron,
                h1e_list, iso=iso, ispin=ispin, ismear=ismear, delta=delta)


def get_total_valence_elec(fname):
    comm = MPI.COMM_WORLD
    myrank = comm.Get_rank()
    if myrank == 0:
        with open(fname, "r") as f:
            for line in f.readlines():
                if "valence charge in whole" in line:
                    n_elec = float(line.split()[-1])
                    break
    else:
        n_elec = None
    n_elec = comm.bcast(n_elec, root=0)
    return n_elec


def wget_symbols_positions(path="./", wprefix="wannier"):
    with open(path+"/"+wprefix+".win", "r") as f:
        symbols = []
        atomic_positions = []
        read_position = 0
        for line in f.readlines():
            if "begin" in line and "atoms_cart" in line:
                read_position = 1
            elif "end" in line and "atoms_cart" in line:
                break
            elif read_position > 0:
                if "bohr" in line.lower():
                    pref = scipy.constant.physical_constants[ \
                            "Bohr radius"]*1.e10
                elif "ang" in line.lower():
                    pref=1.0
                else:
                    line = line.split()
                    _symbol = line[0].split("_")[0]
                    symbol = _symbol[0:1].upper()
                    if len(_symbol) > 1:
                        symbol += _symbol[1:].lower()
                    # remove possible digit
                    for s in symbol:
                        if s.isdigit():
                            symbol = symbol.replace(s,"")
                    symbols.append(symbol)
                    atomic_positions.append(
                            [float(x)*pref for x in line[1:4]])
    return symbols, atomic_positions


def if_gwannier90(corbs_list, delta_charge=0., wpath="./",
        wprefix="wannier", k_grid=None, lrot_list=None,
        iso=1, ispin=1, ismear=0, delta=0.0258,
        icycle=0):
    # read output from wannier90.
    gwannier = w90(wpath, wprefix)
    gmodel = gwannier.model(zero_energy=0.0, min_hopping_norm=1.e-8,
            ignorable_imaginary_part=1.e-8)
    # uniform grid of k-points always contains the origin.
    if k_grid is None:
        k_grid = gwannier.kgrid
    elif isinstance(k_grid, np.int):
        k_grid = np.asarray(gwannier.kgrid)*k_grid
    else:
        assert(len(k_grid)==3), "Dim(k_grid) is not 3!"

    k_mesh = gmodel.k_uniform_mesh(k_grid)
    numk = k_mesh.shape[0]
    wk = 1./numk
    nbmax = gmodel.get_num_orbitals()
    # spin degeneracy
    spin_deg = 3-max(ispin, iso)

    comm = MPI.COMM_WORLD
    myrank = comm.Get_rank()

    # GMPI_x.h5 file
    kvec = set_kvec_para(numk)
    # wannier basis to complex spherical basis tranformation.
    u_wan2csh = get_wann2csh(nbmax, corbs_list)
    h1e_list = get_h1e_list_wannier(gwannier, corbs_list)
    nelectron = 0.
    with h5py.File('BAREHAM_{}.h5'.format(myrank), 'w') as f:
        for ik in range(kvec[0][2],kvec[0][2]+kvec[0][1]):
            kpt = k_mesh[ik]
            hmat = gmodel._gen_ham(kpt)
            # from wannier basis to correlated orbital-ordered csh basis.
            hmat = u_wan2csh.T.conj().dot(hmat).dot(u_wan2csh)
            f['/IKP_{}/ISYM_1/HK0_SPIN1'.format(ik+1)] = hmat.T
            evals, evecs = gmodel._sol_ham(hmat, eig_vectors=True)
            nelectron += np.count_nonzero(evals < 0.)*wk*spin_deg
            f['/IKP_{}/ek0_spin1'.format(ik+1)] = evals
            # evec is actually evec.T
            f['/IKP_{}/T_PSIK0_TO_HK0_BASIS_SPIN1'.format(ik+1)] = \
                    evecs.T.conj()
    nelectron = comm.reduce(nelectron, op=MPI.SUM)
    if myrank == 0:
        if icycle <= 1:
            wrt_ginit(gwannier.symbols, gwannier.lat,
                    gwannier.atomic_positions, u_wan2csh, lrot_list=lrot_list)
        else:
            update_ginit(u_wan2csh)

        ne_list = [[nbmax, 1, nbmax] for k in range(numk)]
        wk_list = [wk for k in range(numk)]
        nelectron = int(nelectron+0.5)+delta_charge
        wrt_gparambands(numk, nbmax, ne_list, wk_list, k_mesh, nelectron,
                h1e_list, iso=iso, ispin=ispin, ismear=ismear, delta=delta)


def set_kvec_para(numk):
    comm = MPI.COMM_WORLD
    ncpu = comm.Get_size()
    nk_per_cpu = numk//ncpu
    if nk_per_cpu*ncpu < numk:
        nk_per_cpu += 1
    myrank = comm.Get_rank()
    kvec = [[]]
    kvec[0].append(myrank)
    k_start = nk_per_cpu*myrank
    kvec[0].append(min(nk_per_cpu, numk-k_start))
    kvec[0].append(k_start)
    with h5py.File("GMPI_{}.h5".format(myrank), "w") as f:
        f["/nvec"] = [1]
        f["/KVEC"] = np.asarray(kvec).T
    return kvec


def get_h1e_list_wannier(gwannier, corbs_list):
    # List of one-body parts of Hamiltonian.
    h1e_list = [[]]
    for i, corbs in enumerate(corbs_list):
        norbs = len(corbs)
        h1e_list[0].append(np.zeros((norbs,norbs), dtype=np.complex))
        for j1 in range(norbs):
            _j1 = corbs[j1]
            for j2 in range(norbs):
                _j2 = corbs[j2]
                h1e_list[0][-1][j1,j2] = gwannier.ham_r[(0,0,0)]["h"][_j1,_j2]\
                        /float(gwannier.ham_r[(0,0,0)]["deg"])
    # Unitary transformation from complex Harmonics to wannier.
    u_csh2wan_list = get_u_csh2wan_all([len(corbs) for corbs in corbs_list])

    for i, u_csh2wan in enumerate(u_csh2wan_list):
        h1e_list[0][i] = u_csh2wan.dot(h1e_list[0][i]).dot(u_csh2wan.T.conj())
    return h1e_list


def get_wann2csh(nbmax, corbs_list):
    '''
    get transformation from wannier basis to complex spherical harmonics basis.
    '''
    # basis transformation matrix which merges the corelated orbitals
    # to the first places, followed by the uncorrelated orbitals.
    ubasis = np.zeros((nbmax,nbmax), dtype=np.complex)
    # the spin-orbital index remapping accordingly.
    orbs_map = []
    for i, corbs in enumerate(corbs_list):
        norbs = len(corbs)
        for j1 in range(norbs):
            orbs_map.append(corbs[j1])
    # appending the uncorrelated orbitals
    for i in range(nbmax):
        if i not in orbs_map:
            orbs_map.append(i)
    # ubasis: <original basis | correlated orbs-first basis>
    for i in range(nbmax):
        ubasis[orbs_map[i],i] = 1.

    # Unitary transformation from complex Harmonics to wannier.
    u_csh2wan_list = get_u_csh2wan_all([len(corbs) for corbs in corbs_list])
    # get the transformation from wannier basis to correlated
    # orbital-ordered complex spherical Harmonics basis.
    u_csh2wan = block_diag(*u_csh2wan_list)
    ncorbs = u_csh2wan.shape[0]
    u_wan2csh = ubasis.copy()
    u_wan2csh[:,:ncorbs] = ubasis[:,:ncorbs].dot(u_csh2wan.T.conj())
    return u_wan2csh


def wrt_ginit(symbols, cell, scaled_positions, u_wan2csh, lrot_list=None):
    with h5py.File("ginit.h5", "w") as f:
        f['/struct/symbols'] = symbols
        f['/struct/cell'] = cell
        f['/struct/scaled_positions'] = scaled_positions
        if lrot_list is not None:
            f['/struct/locrot_list'] = lrot_list
        f["/u_wan2csh"] = u_wan2csh


def update_ginit(u_wan2csh):
    with h5py.File("ginit.h5", "a") as f:
        u = f["/u_wan2csh"][()]
        if u.shape != u_wan2csh.shape:
            del f["/u_wan2csh"]
            f["/u_wan2csh"] = u_wan2csh


def wrt_gparambands(numk, nbmax, ne_list, wk_list, kpoints, nelectron,
        h1e_list, iso=1, ispin=1, ismear=0, delta=0.0258):
    # single file for the dft band structure information
    # beyond that included in 'BAREHAM_$myrank.h5' files.
    with h5py.File('GPARAMBANDS.h5', 'w') as f:
        # spin-orbit coupling
        f['/iso/'] = [iso]
        # spin
        f['/ispin'] = [ispin]
        # k-points dimension
        f['/kptdim'] = [numk]
        # maximal number of bands
        f['/nbmax'] = [nbmax]
        # band indices associated with local orbital construction.
        f['/NE_LIST_SPIN1'] = ne_list
        # k-points weight
        f['/kptwt'] = wk_list
        # k-points
        f["/kpoints"] = kpoints
        # brillouin zone integration method: fermi or gaussian smearing
        f['/ismear'] = [ismear]
        # smearing factor
        f['/delta'] = [delta]
        # number of valence electrons in the wannier manifold
        f['/nelectron'] = [nelectron]
        # number symmetry operations
        f['/symnop'] = [1]
        # the index of identity symmetry operation
        f['/symie'] = [1]
        for isp, h1es in enumerate(h1e_list):
            for i, h1e in enumerate(h1es):
                # local one-body of the dft hamiltonian
                f['/IMPURITY_{}/H1E_SPIN{}'.format(i+1, isp+1)] = h1e.T


def wannier_den_matrix(wannier_path="./"):
    '''produce the file `wannier_den_matrix.dat` for the feedback from
    g-risb to dft.
    '''
    _, _, kpts, include_bands, wfwannier_list, bnd_es_in = \
            mpiget_wannier_data(path=wannier_path)
    bnd_es, bnd_vs = mpiget_bndev(kpts, wfwannier_list=wfwannier_list,
            bnd_es_in=bnd_es_in, mode="risb")
    nktot = len(kpts)
    with h5py.File("GPARAMBANDS.h5", "r") as f:
        delta = f["/delta"][0]
        ismear = f["/ismear"][0]
        iso = f["/iso"][0]
        num_elec = f["/nelectron"][0]
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    # set wk_list
    wklist = [1./nktot for i in range(bnd_es.shape[1])]
    if rank == 0:
        efermi = get_fermi_level(bnd_es, wklist, num_elec, delta=delta, \
                ismear=ismear, iso=iso)
    else:
        efermi = None
    efermi = comm.bcast(efermi, root=0)
    ncpu = comm.Get_size()
    nk_cpu = nktot//ncpu
    if nk_cpu*ncpu < nktot:
        nk_cpu += 1
    # reduce bnd_es to local part only
    if rank == 0:
        bnd_es = bnd_es[:nk_cpu]
        wklist = wklist[:nk_cpu]
    # set fermi weight
    ferwes = get_fermi_weight(efermi, bnd_es, wklist, delta=delta,
            ismear=ismear, iso=iso)
    # calculate wannier_den_matrix
    wan_den = get_wannier_den_matrix_risb(bnd_vs, ferwes, wklist, nktot)
    if rank == 0:
        wrt_wan_den(wan_den, wfwannier_list, include_bands,
                fwannier="{}/wannier.dat".format(wannier_path))


def wannier_den_matrix_lda_chk(wannier_path="./"):
    '''produce the file `wannier_den_matrix.dat` for the feedback from
    g-risb to dft.
    '''
    _, _, kpts, include_bands, _, bnd_es = mpiget_wannier_data(
            path=wannier_path)
    nktot = len(kpts)
    with h5py.File("GPARAMBANDS.h5", "r") as f:
        num_elec = f["/nelectron"][0]

    # chop bnd_es
    nbnd = bnd_es.shape[2]
    nwan = int(num_elec/2+3)
    bnd_es = bnd_es[:,:,:nwan]
    # set wk_list
    wklist = [1./nktot for i in range(bnd_es.shape[1])]
    efermi = get_fermi_level(bnd_es, wklist, num_elec, ismear=-1)
    # set fermi weight
    ferwes = get_fermi_weight(efermi, bnd_es, wklist, ismear=-1)
    # setup trivial wannier_den data.
    wan_den = [[]]
    wfwannier_list = [[]]
    vmat = np.zeros((nbnd, nwan), dtype=np.complex)
    np.fill_diagonal(vmat, 1.0)
    for ik in range(nktot):
        wfwannier_list[-1].append(vmat)
        wan_den[0].append(np.diag(ferwes[0][ik]*nktot/(2+0.j)))
    wan_den = np.asarray(wan_den)
    wfwannier_list = np.asarray(wfwannier_list)
    wrt_wan_den(wan_den, wfwannier_list, include_bands,
            fwannier="{}/wannier.dat".format(wannier_path))


def wannier_den_matrix_lda_chk2(wannier_path="./"):
    '''produce the file `wannier_den_matrix.dat` for the feedback from
    g-risb to dft.
    '''
    _, _, kpts, include_bands, wfwannier_list, bnd_es = mpiget_wannier_data(
            path=wannier_path)
    nktot = len(kpts)
    with h5py.File("GPARAMBANDS.h5", "r") as f:
        num_elec = f["/nelectron"][0]
    # set wk_list
    wklist = [1./nktot for i in range(bnd_es.shape[1])]
    efermi = get_fermi_level(bnd_es, wklist, num_elec, ismear=-1)
    # set fermi weight
    ferwes = get_fermi_weight(efermi, bnd_es, wklist, ismear=-1)
    # setup trivial wannier_den data.
    wan_den = [[]]
    for ik in range(nktot):
        dm = np.diag(ferwes[0][ik]*nktot/(2+0.j))
        wan_den[0].append(wfwannier_list[0][ik].T.conj().dot(dm).dot(
            wfwannier_list[0][ik]))
    wan_den = np.asarray(wan_den)
    wrt_wan_den(wan_den, wfwannier_list, include_bands,
            fwannier="{}/wannier.dat".format(wannier_path))


def wannier_den_matrix_lda_chk3(wannier_path="./"):
    '''produce the file `wannier_den_matrix.dat` for the feedback from
    g-risb to dft.
    '''
    _, _, kpts, include_bands, wfwannier_list, bnd_es = mpiget_wannier_data(
            path=wannier_path)
    nktot = len(kpts)
    with h5py.File("GPARAMBANDS.h5", "r") as f:
        num_elec = f["/nelectron"][0]
    # set wk_list
    wklist = [1./nktot for i in range(bnd_es.shape[1])]
    # get eigen-vector from interpolation
    bnd_es2 = [[]]
    bnd_ev2 = [[]]
    for ik in range(nktot):
        hk = wfwannier_list[0][ik].T.conj().dot(np.diag(bnd_es[0][ik])).dot(
                wfwannier_list[0][ik])
        w, v = np.linalg.eigh(hk)
        bnd_es2[0].append(w)
        bnd_ev2[0].append(v)
    bnd_ev2 = np.asarray(bnd_ev2)
    with h5py.File("ev_lda_ref.h5", "w") as f:
            f["e"] = bnd_es2
            f["v"] = bnd_ev2
    efermi = get_fermi_level(bnd_es2, wklist, num_elec, ismear=-1)
    # set fermi weight
    ferwes = get_fermi_weight(efermi, bnd_es2, wklist, ismear=-1)
    # setup trivial wannier_den data.
    wan_den = [[]]
    for ik in range(nktot):
        dm = np.diag(ferwes[0][ik]*nktot/(2+0.j))
        wan_den[0].append(bnd_ev2[0][ik].dot(dm).dot(
            bnd_ev2[0][ik].T.conj()))
    wan_den = np.asarray(wan_den)
    wrt_wan_den(wan_den, wfwannier_list, include_bands,
            fwannier="{}/wannier.dat".format(wannier_path))


def wrt_wan_den(wan_den, wfwannier_list, include_bands, fwannier=None):
    try:
        with h5py.File(fwannier, "r") as f:
            h5wrt_wan_den(wan_den, wfwannier_list, include_bands)
    except:
        fwrt_wan_den(wan_den, wfwannier_list, include_bands)


def fwrt_wan_den(wan_den, wfwannier_list, include_bands):
    """write wannier_den_matrix.dat file in fortran binary format.
    for spin-restricted case only as of now.
    note the index order wfwannier_list[isp, ibnd, iwann].
    """
    nktot = wan_den.shape[1]
    wfwannier_list = wfwannier_list.swapaxes(2, 3)
    nband = wfwannier_list.shape[3]
    nwann = wfwannier_list.shape[2]
    with FortranFile('wannier_den_matrix.dat', 'w') as f:
        f.write_record(300.0)
        f.write_record(nktot)
        f.write_record([nband for k in range(nktot)])
        f.write_record([nwann for k in range(nktot)])
        f.write_record([include_bands for k in range(nktot)])
        f.write_record(wfwannier_list[0])
        f.write_record(wan_den[0].swapaxes(1, 2))


def h5wrt_wan_den(wan_den, wfwannier_list, include_bands):
    """write wannier_den_matrix.dat file in hdf5 format.
    for spin-restricted case only as of now.
    note the index order wfwannier_list[isp, ibnd, iwann].
    """
    nktot = wan_den.shape[1]
    wfwannier_list = wfwannier_list.swapaxes(2, 3)
    nband = wfwannier_list.shape[3]
    nwann = wfwannier_list.shape[2]
    with h5py.File('wannier_den_matrix.dat', 'w') as f:
        f["/temperature"] = [300.0]
        f["/nqdiv"] = [nktot]
        f["/num_band_in"] = [nband for k in range(nktot)]
        f["/n_dmft_wan"] = [nwann for k in range(nktot)]
        f["/include_band_new"] = [include_bands for k in range(nktot)]
        f["/vw_matrix_new"] = wfwannier_list[0]
        f["/n_matrix_new"] = wan_den[0].swapaxes(1, 2)


def get_risb_bndes(path="./"):
    num_list = [int(x.split("_")[1].split(".")[0]) \
            for x in glob.glob(path+"/GBANDS_*.h5")]
    num_list.sort()
    bnd_es = []
    for isp in range(2):
        for i in num_list:
            fband = "GBANDS_{}.h5".format(i)
            with h5py.File(fband, "r") as f:
                if "/ISPIN_{}".format(isp+1) in f:
                    if len(bnd_es) == isp:
                        bnd_es.append([])
                    ik_start = f["/IKP_START"][0]
                    ik_end = f["/IKP_END"][0]
                    for ik in range(ik_start, ik_end+1):
                        bnd_es[isp].append(f["/ISPIN_{}/IKP_{}/ek".format(\
                                isp+1, ik)][()])
    bnd_es = np.asarray(bnd_es)
    return bnd_es



from __future__ import print_function

"""
Atom Symmetry module

Input: 3D rotation group and a representation of J = [J_{x}, J_{y}, J_{z}]
of an atom

Output:

1) A new representation of Jp and a unitary transformation U such that
   Jp = U^H J U
2) A symbolic matrix representing a general matrix commuting with
   the representation of the group.
"""

import numpy as np
from scipy.sparse import isspmatrix
import sys


def get_3D_rotation_generator():
    return -1.j * np.array([[[int((i - j) * (j - k) * (k - i) / 2)
        for k in range(3)] for j in range(3)] for i in range(3)])


def get_rotation(J, theta, method="eigh"):
    '''
    Get rotation matrix by Lie parameters and generator J.
    '''
    exponent = J[0] * theta[0] + J[1] * theta[1] + J[2] * theta[2]
    if method == "eigh":
        if isspmatrix(exponent):
            exponent = exponent.todense()
        vals, vecs = np.linalg.eigh(exponent)
        rep = np.dot(
            np.dot(vecs, np.diag(np.exp(1.j * vals))), np.conj(vecs).T)
    elif method == "expm":
        from scipy.linalg import expm
        rep = expm(1.j * exponent)
    else:
        raise AssertionError(
                " method =" + method + ": Unsupported in get_rotation!")
    return rep


def get_Lie_parameter_3D(g, plus2pi=False):
    '''
    Get Lie parameter from a 3D rotation matrix.
    '''
    J = get_3D_rotation_generator()
    list_theta = []
    assert len(g) == 3, \
            ' Error: get_Lie_parameter_3D is exclusively for 3D rotations!'

    # all rotation matrices are diagonalizable over the complex field
    vals, vecs = np.linalg.eig(g)

    # pick the label of the eigenvalue 1 (rotation axis)
    rotation_axes_label = np.abs(
        vals - 1.0).tolist().index(min(np.abs(vals - 1.0)))
    axis = vecs[:, rotation_axes_label].real
    fix = (np.trace(g) - 1.0) / 2.0
    if fix <= -1.0:
        assert(np.abs((np.trace(g) - 1.0) / 2.0 - (-1.0)) < 1.e-10)
        fix = -1.0
    if fix >= 1.0:
        assert(np.abs((np.trace(g) - 1.0) / 2.0 - (1.0)) < 1.e-10)
        fix = 1.0
    alpha = np.arccos(fix)
    theta = alpha * axis

    # get the additional sign
    gp = get_rotation(J, theta, method="eigh")
    err = np.max(np.abs(gp - g))
    if err > 1.e-7:
        alpha = -alpha
    list_theta.append(alpha * axis)
    if plus2pi:
        list_theta.append((alpha + np.pi * 2.) * axis)
    return list_theta


def get_Lie_parameters(Rotations, plus2pi=False):
    '''
    Get Lie parameters for a 3D rotation group.
    '''
    Lie_parameters = []
    for g in Rotations:
        list_theta = get_Lie_parameter_3D(g, plus2pi=plus2pi)
        for theta in list_theta:
            Lie_parameters.append(theta)
    return Lie_parameters


def get_plus2pi(Jz):
    '''
    Decide whether one needs to double the number of Lie parameters
    due to half-integer of J.
    '''
    vals = np.linalg.eigvalsh(Jz)
    val = np.min(np.abs(vals))
    if np.abs(val - 0.5) < 1.e-7:
        plus2pi = True
    elif val < 1.e-7:
        plus2pi = False
    else:
        print(' eigen-vales of Jz:\n {}'.format(vals))
        raise AssertionError('Error in eigen-values of Jz!')
    return plus2pi


def check_Lie_parameters(G, J, Lie_parameters):
    '''
    Check the representation generated by Lie parameters and J with G.
    '''
    N = len(Lie_parameters) // len(G)  # can be 1 or 2
    for i, g in enumerate(G):
        for j in range(i * N, (i + 1) * N):
            theta = Lie_parameters[j]
            gp = get_rotation(J, theta, method="eigh")
            err = np.max(np.abs(gp - g))
            assert err < 1.e-7, " check_Lie_parameters!"


def check_Lie_parameters_3D(Rotations, Lie_parameters):
    '''
    Compare the 3D rotation matrices with the one generated
    by Lie parameters and the generator J.
    '''
    J = get_3D_rotation_generator()
    check_Lie_parameters(Rotations, J, Lie_parameters)


def get_ratio_matab(a, b):
    '''
    Get a scale ratio of two matrix.
    '''
    ratio = 0. + 0.j
    # locate the first significant element of a
    ia = -1
    ja = -1
    import itertools
    for i, j in itertools.product(range(len(a)), repeat=2):
        if np.abs(a[i][j]) > 1.e-5:
            ia = i
            ja = j
            break
    if ia >= 0:
        if np.abs(b[ia][ja]) > 1.e-7:
            rat = a[ia, ja] / b[ia, ja]
            if(np.max(np.abs(a - b * rat)) < 1.e-6):
                ratio = rat
    return ratio


def get_structure_constants(J):
    '''
    Get the structure constant or commutation table of [Jx,Jy,Jz].
    '''
    l = len(J)
    eijk = np.zeros((l, l, l), dtype=np.complex)
    for i0, J0 in enumerate(J):
        for i1, J1 in enumerate(J):
            J01 = np.dot(J0, J1) - np.dot(J1, J0)
            for i2, J2 in enumerate(J):
                eijk[i0, i1, i2] = get_ratio_matab(J01, J2)
    return eijk


def check_strcture_constants(J, Jp):
    '''
    Compare the structure constants between two sets of generators.
    '''
    eijk = get_structure_constants(J)
    eijkp = get_structure_constants(Jp)
    if np.max(np.abs(eijkp - eijk)) > 1.e-6:
        print(' Commutation table for J:')
        print(eijk)
        print(' Commutation table for Jp:')
        print(eijkp)
        raise AssertionError('Inconsistency error!')


def check_strcture_constants_3D(J):
    Jp = get_3D_rotation_generator()
    check_strcture_constants(J, Jp)


def get_representation(J, Lie_parameters):
    representation = []
    for theta in Lie_parameters:
        repr = get_rotation(J, theta, method="eigh")
        try:
            # try removing the small elements.
            repr = repr.multiply(np.abs(repr) > 1.e-8)
        except:
            pass
        representation.append(repr)
    return representation


def locate_R_list(R01, R_list):
    '''
    Locate the R in R_list.
    '''
    loc = -1
    min_dist = 1.e5
    for i, R in enumerate(R_list):
        dist = np.max(np.abs(R01 - R))
        if (dist < min_dist):
            min_dist = dist
            if(dist < 1.e-6):
                if loc == -1:
                    loc = i
                else:
                    raise AssertionError(
                            'Found multiple identical elements!')
    if loc == -1:
        print(" Rotation operations list:")
        print_R_list(R_list)
        raise AssertionError(
                "Failed to locate the element with min_dist =" +\
                " %12.3f"%(min_dist))
    return loc


def print_R_list(R_list):
    for i, R in enumerate(R_list):
       print("%3d %5.2f %5.2f %5.2f"%((i,)+tuple(R[0])))
       print("    %5.2f %5.2f %5.2f"%tuple(R[1]))
       print("    %5.2f %5.2f %5.2f"%tuple(R[2]))


def get_product_table(R_list):
    '''
    Get product table of the representation.
    '''
    h = len(R_list)
    table = np.zeros((h, h), dtype=np.int)
    for i0, R0 in enumerate(R_list):
        for i1, R1 in enumerate(R_list):
            R01 = np.dot(R0, R1)
            table[i0, i1] = locate_R_list(R01, R_list)
    return table


def check_product_table(G, representation):
    product_table_G = get_product_table(G)
    product_table_R = get_product_table(representation)
    if len(G) == len(representation):
        err = np.max(np.abs(product_table_G - product_table_R))
        if err > 1.e-6:
            raise AssertionError('Product table not match!')


def get_eigen_space(H, l_list, tol=1.e-6):
    '''
    Get eigen spaces of a Hermitian matrix H,
    in accordance with the possible block form specified by l_list.
    '''
    from scipy.linalg import block_diag
    val_list = []
    vec_list = []
    lbase = 0
    for l1 in l_list:
        vals, vecs = np.linalg.eigh(H[lbase:lbase + l1, lbase:lbase + l1])
        lbase += l1
        val_list += vals.tolist()
        vec_list.append(vecs)
    vals = np.array(val_list)
    vecs = block_diag(*vec_list)
    list_eigen_space = []
    val = vals[0]
    eigen_space = []
    N = len(vals)
    for i in range(N):
        if np.abs((vals[i] - val) / val) < tol:
            eigen_space.append(vecs[:, i])
        else:
            # note eigen_space in c-convention
            list_eigen_space.append(np.array(eigen_space))
            eigen_space = []
            eigen_space.append(vecs[:, i])
            val = vals[i]
    list_eigen_space.append(np.array(eigen_space))
    return list_eigen_space


def brief_eigen_spaces(list_eigen_space):
    print(' Number of eigen space =', len(list_eigen_space))
    print(' Dimension of each eigen space:')
    for i, eigen_space in enumerate(list_eigen_space):
        print(i, len(eigen_space))


def check_commute_G(H, G):
    for i, g in enumerate(G):
        if isinstance(g, np.ndarray):
            C = np.dot(H, g) - np.dot(g, H)
        elif isspmatrix(g):
            C = np.dot(H, g.todense()) - g.dot(H)
        else:
            raise AssertionError("Unsupported datatype of g" + g.__class__)
        err = np.max(np.abs(C))
        if err > 1.e-6:
            raise AssertionError("Error in check_commute_G = %12.7f"%(err))


def get_characters_espace(representation, V):
    '''
    Get the character (vector) of one eigen-space defined by V.
    '''
    chi = []
    rep = []
    for R in representation:
        # note here V in c-convention
        VHRV = V.conj().dot(R).dot(V.T)
        if VHRV.shape == (1,):
            tr = VHRV[0]
        else:
            tr = np.trace(VHRV)
        chi.append(tr)
        rep.append(VHRV)
    return np.array(chi), rep


def get_sub_repr_characters(representation, list_eigen_space):
    list_chi = []
    list_repr = []
    for i, V in enumerate(list_eigen_space):
        chi, rep = get_characters_espace(representation, V)
        list_chi.append(chi)
        list_repr.append(rep)
    return np.array(list_chi), list_repr


def check_sum_chi2_1(chi):
    irr = 'OK'
    h = len(chi)  # order of the group
    sum_chi2 = np.vdot(chi, chi).real
    if np.abs(sum_chi2 - h) > 1.e-5:
        print(' Error in sum_chi2 = ' + str(np.abs(sum_chi2 - h)))
        irr = 'NO'
    return irr


def check_sum_chi2(list_chi):
    '''
    Check sum of character^2
    '''
    for i, chi in enumerate(list_chi):
        irr = check_sum_chi2_1(chi)
        if irr == 'NO':
            return irr
    return irr


def get_ired_representation(representation):
    '''get irreducible representations.
    '''
    l = representation[0].shape[0]
    from pygrisb.math.matrix_util import get_matrices_bk_list, sym_array
    from scipy.linalg import block_diag
    check = 'no'
    if '-allmix' in sys.argv:
        l_list = [l]
    else:
        l_list = get_matrices_bk_list(representation)
    print(' list of (sub)dimensions of the representation:\n {}'\
            .format(l_list))

    while check == 'no':
        X = []
        for l1 in l_list:
            X.append(np.random.rand(l1, l1) + np.random.rand(l1, l1) * 1.j)
        X = block_diag(*X)
        # make X Hermitian
        X += X.conj().T
        H = sym_array(X, representation)
        check_commute_G(H, representation)
        list_eigen_space = get_eigen_space(H, l_list)
        list_chi, list_repr = get_sub_repr_characters(
            representation, list_eigen_space)
        check = check_sum_chi2(list_chi)
        if check == 'ok':
            break
    return list_eigen_space, list_chi, list_repr


def get_chi_repr(list_chi, list_repr, list_eigen_space):
    '''
    Group the equivalent (chi) representations.
    '''
    list_chi_repr = []
    list_chi_space = []
    list_distinct_chi = []
    done_list = []
    h = len(list_chi[0])
    chi_product = np.dot(np.conj(list_chi), list_chi.T) / h

    for i, product in enumerate(chi_product):
        chi_repr = []
        chi_space = []
        if i in done_list:
            continue
        for j in range(len(product)):
            if np.abs(product[j] - 1) < 1.e-6:
                if j < i:
                    raise AssertionError('Error in get_chi_repr!')
                chi_repr.append(list_repr[j])
                chi_space.append(list_eigen_space[j])
                done_list.append(j)
                if len(chi_repr) == 1:
                    list_distinct_chi.append(list_chi[j])
        list_chi_repr.append(chi_repr)
        list_chi_space.append(np.asarray(chi_space))
        print( ' chi_space {} dimension = {}'.format(len(list_chi_space) - 1,\
                np.asarray(chi_space).shape))

    return list_chi_repr, list_chi_space, list_distinct_chi


def get_U_equiv_repr(representationA, representationB):
    '''
    Get the transformation matrix U such that B= UAU^H.
    '''
    M = np.zeros(representationA[0].shape)
    for i in range(len(representationA)):
        M = M + representationB[i].dot(representationA[i].T.conj())
    MM = M.dot(M.T.conj())
    vals, vecs = np.linalg.eigh(MM)
    MM = vecs.dot(np.diag(np.sqrt(vals)**(-1))).dot(vecs.T.conj())
    # order is important!
    utrans = MM.dot(M)
    return utrans


def check_U_equiv_repr(representationA, representationB, U):
    for i in range(len(representationA)):
        UAUH = np.dot(U, np.dot(representationA[i], np.conj(U).T))
        err = np.max(np.abs(UAUH - representationB[i]))
        if err > 1.e-5:  # 10:
            raise AssertionError(
                    'Large error in check_U_equiv_repr = %10.5f'%(err))


def get_list_U_equiv_repr(list_chi_repr):
    list_U = []
    for chi_repr in list_chi_repr:
        U = []
        for i in range(1, len(chi_repr)):
            U1 = get_U_equiv_repr(chi_repr[i], chi_repr[0])
            check_U_equiv_repr(chi_repr[i], chi_repr[0], U1)
            U.append(U1)
        list_U.append(U)
    return list_U


def get_U_Jnew(list_chi_space, list_chi_U, reorder=True):
    '''
    Reorder basis to get nice block form.
    '''
    U_Jnew = []
    for i, chi_space in enumerate(list_chi_space):
        for j in range(1, len(chi_space)):
            # Fortran c row/column convention different
            chi_space[j] = np.dot(np.conj(list_chi_U[i][j - 1]), chi_space[j])
        if reorder:
            for k in range(len(chi_space[0])):
                for j in range(len(chi_space)):
                    U_Jnew.append(chi_space[j][k])
        else:
            for j in range(len(chi_space)):
                for k in range(len(chi_space[0])):
                    U_Jnew.append(chi_space[j][k])
    # back to Fortran convention
    U_Jnew = np.array(U_Jnew).T
    print(" U_Jnew unitary error = {}".format(np.max(np.abs(
            U_Jnew.T.conj().dot(U_Jnew)- np.eye(U_Jnew.shape[0])))))
    return U_Jnew


def get_seq_matrix(N):
    '''
    Generate a square matrix with sequential elements.
    '''
    M = []
    for i in range(N):
        Mi = []
        for j in range(N):
            Mi.append(j + 1 + i * N)
        M.append(Mi)
    return np.array(M)


def get_sigma_struct(list_chi_space):
    sigma = []
    ibase = 0
    for chi_space in list_chi_space:
        # number of equivalent irred. repr.
        N = len(chi_space)
        L = len(chi_space[0])
        M = get_seq_matrix(N) + ibase
        for i in range(L):
            sigma.append(M)
        ibase = np.max(M)
    return sigma


def get_Lie_from_JR(Rotations, J):
    '''
    Get the Lie parameters of {R}.
    '''
    plus2pi = get_plus2pi(J[2])
    Lie_parameters = get_Lie_parameters(Rotations, plus2pi=plus2pi)
    check_Lie_parameters_3D(Rotations, Lie_parameters)
    return Lie_parameters


def get_JR_representation(Rotations, J):
    '''
    Get the representation of {R} using generator {J}.
    '''
    Lie_parameters = get_Lie_from_JR(Rotations, J)
    check_strcture_constants_3D(J)
    representation = get_representation(J, Lie_parameters)
    return representation


def get_diag_sub_array(H, I1, I2, J1, J2):
    err = 0.
    N = len(H)
    if I1 > 0:
        err = np.max(err, np.max(np.abs(H[:I1, J1:J2])))
    if I2 < N:
        err = np.max(err, np.max(np.abs(H[I2:, J1:J2])))
    if J1 > 0:
        err = np.max(err, np.max(np.abs(H[I1:I2, :J1])))
    if J2 < N:
        err = np.max(err, np.max(np.abs(H[I1:I2, J2:])))
    assert err < 1.e-7, ' Subarray not diagonal, err > 1.e-7!'
    return H[I1:I2, J1:J2]


def get_atom_U_sigma_chi(representation):
    list_eigen_space, list_chi, list_repr = get_ired_representation(
        representation)
    list_chi_repr, list_chi_space, list_distinct_chi = get_chi_repr(
        list_chi, list_repr, list_eigen_space)
    list_chi_U = get_list_U_equiv_repr(list_chi_repr)
    U = get_U_Jnew(list_chi_space, list_chi_U)
    sigma = get_sigma_struct(list_chi_space)
    return U, sigma, list_distinct_chi


def get_atom_Jnew(Rotations, J):
    '''
    Main function.
    '''
    print(" Rotations list:")
    print_R_list(Rotations)
    representation = get_JR_representation(Rotations, J)
    check_product_table(Rotations, representation)
    U, sigma, chi = get_atom_U_sigma_chi(representation)
    from scipy.linalg import block_diag
    Jnew = []
    for J1 in J:
        Jnew.append(U.conj().T.dot(J1).dot(U))
    return Jnew, U, block_diag(*sigma)


def check_commute_sigma_J(sigma, J, Rotations):
    '''
    Check the commutation of sigma with the new rotated representation.
    '''
    representation = get_JR_representation(Rotations, J)
    check_commute_G(sigma, representation)



import os, sys, subprocess
from pygrisb.gutz import embedh


def driver_c2rspci_mott_onfly(imp=1):
    '''exe_spci_mott_onfly with complex embedding Hamiltonian to
    real transformation.
    '''
    if "-i" in sys.argv:
        imp = int(sys.argv[sys.argv.index("-i")+1])
    if os.path.isfile("EMBED_HAMIL_{}r.h5".format(imp)):
        lv2e = False
    else:
        lv2e = True
    # get embedding hamiltonian parameters.
    h1e, lambdac, daalpha, v2e = embedh.h5gen_embedh_spin_updn(imp=imp,
            lv2e=lv2e)
    # write real version of the embedding hamiltonian.
    embedh.h5wrt_rembed_hamil(h1e, lambdac, daalpha, v2e, imp=imp)

    cmd = [os.environ["WIEN_GUTZ_ROOT2"]+"/exe_rspci_mott_onfly", str(imp)]
    subprocess.call(cmd)

    # get results
    dm, e_mol = embedh.get_res_rspci_mott_onfly(imp=imp)
    embedh.h5wrt_dm_sab_rc(dm, e_mol, imp=imp)


if __name__ == "__main__":
    driver_c2rspci_mott_onfly()



import os, sys, h5py
import numpy as np
from pygrisb.model import checkboard
import subprocess

import matplotlib
matplotlib.rcParams['backend'] = "Qt5Agg"
import matplotlib.pyplot as plt


def generate_data(u_list, spindeg=True, fname='result', iembeddiag=-1):
    '''run *CyGutz* calculations for a list of U.

    Parameters:

    * u_list: real array
      list of Hubbard U parameters.
    * spindeg: boolean number
      whether to keep spin degeneracy or not.
    * fname: string
      file name to store the results.
    * iembeddiag: integer
      flag for method to solve the embedding Hamiltonian.

      * -3: valence truncation ED with S=0 (spin-singlet) constraint;
      * -1: valence truncation ED;
      * 10: Hartree-Fock.

    Result:

    save the list of energies (e_list), double occupancy (d_list),
    quasi-particle weight (z_list), magnetic moment (m_list),
    to ``fname.dat`` text file as well as the ``fname.h5`` hdf5 file.
    '''
    # get *CyGutz* command. Choose option '-r -1' to skip the claculation of
    # Gutzwiller renormalized electron density.
    root = os.environ['WIEN_GUTZ_ROOT2']
    cmd = [root+'/CyGutz', '-r', '-1', '-n', '500']

    # set Hubbard U=0
    u = 0.

    # remove pre-existing Gutzwiller setup files.
    for f in ['ginit.h5', 'GPARAM.h5']:
        if os.path.exists(f):
            os.remove(f)

    # generate input files with updated with u=0
    checkboard.gutz_model_setup(u=u, spindeg=spindeg, iembeddiag=iembeddiag)

    # total energy list
    e_list = []

    # quasi-particle weight z list
    z_list = []

    # double occupancy d list
    d_list = []

    # magnetic moment
    m_list = []

    # loop over the list of U.
    for u in u_list:

        print(' working on u = {}'.format(u))

        # modify the local Coulomb matrix of two sites
        with h5py.File('GPARAM.h5', 'a') as f:
            for i in range(1,3):
                # note the transposition, which is transformation
                # from Fortran convention to c-convention.
                v2e = f['/IMPURITY_{}/V2E'.format(i)][()].T

                # now update the Coulom matrix
                v2e[0,0,0,0] = v2e[0,0,1,1] = v2e[1,1,0,0] = v2e[1,1,1,1] = u
                f['/IMPURITY_{}/V2E'.format(i)][()] = v2e.T

            f['/dc_u_avg'][()] = [u, u]

        # perform the *CyGutz* calculation.
        subprocess.call(cmd)

        # get total energy
        with h5py.File('GLOG.h5', 'r') as f:
            e = f['/etot_model'][0]
            e_list.append(e.real)

        # get Z = R^\dagger R
        with h5py.File('WH_RL_OUT.h5', 'r') as f:
            r = f['/IMPURITY_1/R'][0,0]

            # simple here since it is just a scalar
            z = r*r.conj()
            z_list.append(z.real)

        # if unrestricted HF calculations--
        if iembeddiag == 10:
            # in the single-orbital case, double occupancy is simply
            # <n_up>_0 <n_dn>_0
            with h5py.File('GLOG.h5', 'r') as f:
                nup = f['/IMPURITY_1/NKS_SYM'][0,0]
                ndn = f['/IMPURITY_1/NKS_SYM'][1,1]
                d = nup*ndn
                d_list.append(d.real)
                m = nup - ndn
                m_list.append(m)
        else:
            # to get double occupancy (of impurity 1), <n_up n_dn>_G,
            # we run analysis code *exe_spci_analysis*
            subprocess.call([root+'/exe_spci_analysis', '1'])

            # double occupancy is simply the local many-body
            # density matrix element in the valence=2 block.
            with h5py.File('EMBED_HAMIL_ANALYSIS_1.h5', 'r') as f:
                if '/valence_block_2/RHO' in f:
                    d = f['/valence_block_2/RHO'][0, 0]
                else:
                    d = 0.
                d_list.append(d.real)

            with h5py.File('EMBED_HAMIL_RES_1.h5', 'r') as f:
                # get the density matrix of the embedding Hamiltonian
                # note the first half of the orbotals correspond to
                # the physical one-body space
                nup = f['/DM'][0, 0]
                ndn = f['/DM'][1, 1]
                m = nup - ndn
                m_list.append(m.real)

    # save to text file.
    with open(fname+'.dat', 'w') as f:
        for u, e, z, d, m in zip(u_list, e_list, z_list, d_list, m_list):
            f.write('{} {} {} {} {}\n'.format(u, e, z, d, m))

    # save to hdf5 file.
    with h5py.File(fname+'.h5', 'w') as f:
        f['/u_list'] = u_list
        f['/e_list'] = e_list
        f['/z_list'] = z_list
        f['/d_list'] = d_list
        f['/m_list'] = m_list


def scan_u(spindeg=True, iembeddiag=-1, fname='result'):
    '''run *CyGutz* calculations for a list of U.

    Result:

    it will generate results for a u_list of np.arange(0.0, 5.1, 0.2).
    '''
    if os.path.isfile(fname+'.h5'):
        return

    # set range of Hubbard U.
    u_list = np.arange(0.0, 20.0, 0.5)
    generate_data(u_list, spindeg=spindeg, fname=fname, iembeddiag=iembeddiag)


def get_scan_data(fname='result'):
    with h5py.File(fname+'.h5', 'r') as f:
        u_list = f['/u_list'][()]
        e_list = f['/e_list'][()]
        z_list = f['/z_list'][()]
        d_list = f['/d_list'][()]
        m_list = f['/m_list'][()]

    return u_list, e_list, z_list, d_list, m_list


def plot_scan_u(fname='result'):
    u_list, e_list, z_list, d_list, m_list = get_scan_data(fname=fname)

    f, axarr = plt.subplots(2, 2, sharex=True)
    axarr[0, 0].plot(u_list, e_list)
    axarr[0, 0].set_ylabel('energy')
    axarr[1, 0].plot(u_list, d_list)
    axarr[1, 0].set_ylabel('double occupancy')
    axarr[0, 1].plot(u_list, z_list)
    axarr[0, 1].yaxis.tick_right()
    axarr[0, 1].yaxis.set_label_position("right")
    axarr[0, 1].set_ylabel('Z')
    axarr[1, 1].plot(u_list, m_list)
    axarr[1, 1].set_ylabel('local $<S_{z}>$')
    axarr[1, 1].set_ylim(-1,1)
    axarr[1, 1].yaxis.tick_right()
    axarr[1, 1].yaxis.set_label_position("right")
    axarr[1, 0].set_xlabel('U')
    axarr[1, 1].set_xlabel('U')
    axarr[1, 0].set_xlim(min(u_list), max(u_list))
    plt.tight_layout()
    plt.show()
    f.savefig(fname+'.png')


if __name__=='__main__':
    fname=''
    if '-sp' in sys.argv:
        spindeg = False
        fname = 'result_afm'
        iembeddiag = -1
        scan_u(spindeg=False, fname='result_afm')
    elif '-uhf' in sys.argv:
        spindeg = False
        fname = 'result_afm_uhf'
        iembeddiag = 10
    elif '-rhf' in sys.argv:
        spindeg = True
        fname = 'result_pm_rhf'
        iembeddiag = 10
    else:
        spindeg = True
        fname = 'result_pm'
        iembeddiag = -3

    scan_u(spindeg=spindeg, fname=fname, iembeddiag=iembeddiag)
    plot_scan_u(fname=fname)



import h5py
import numpy as np


def gh5_produce_random_rl(norb, l, z, h):

    norb2 = norb*2

    # R matrix
    r_mat = np.zeros((norb, norb), dtype=complex)
    r_mat[0,0] = z
    r_mat[1,0] = h
    r_mat[2,0] = h
    r_mat2 = np.zeros((norb2, norb2), dtype=complex)
    r_mat2[0::2, 0::2] = r_mat
    r_mat2[1::2, 1::2] = r_mat2[0::2, 0::2]

    # \lambda matrix
    l_mat = np.zeros((norb, norb), dtype=complex)
    l_mat[1,1] = l
    l_mat[2,2] = -l
    l_mat2 = np.zeros((norb2, norb2), dtype=complex)
    l_mat2[0::2, 0::2] = l_mat
    l_mat2[1::2, 1::2] = l_mat2[0::2, 0::2]

    with h5py.File('WH_RL_INIT.h5', 'w') as f:
        f['/IMPURITY_1/LAMBDA'] = l_mat2.T
        f['/IMPURITY_1/R'] = r_mat2.T


if __name__=='__main__':
    gh5_produce_random_rl(norb=3, l=1.1239, z=0.407591572, h=0.6459494)



import numpy as np
from ase.dft import kpoints
import pygrisb.gutz.ginput as ginput
import pygrisb.model.tbASE as tb
import scipy.constants as s_const

# The following is a simple test for the 1-d Hubbard model.

a = tb.AtomsTB("N", [(0, 0, 0)], cell=(1, 1, 1))
a.set_orbitals_spindeg(orbitals=[("s","p")])
aTB = tb.TB(a)
aTB.set_hop([
        (( 1,0,0),0,0,-1),
        ((-1,0,0),0,0,-1),
        (( 1,0,0),1,1,-1),
        ((-1,0,0),1,1,-1),
        ((0,0,0),0,1,0.25),
        ((0,0,0),1,0,0.25)])

kps_size = (100, 1, 1)
kps = kpoints.monkhorst_pack(kps_size)

num_k = len(kps)
kps_wt = 1.0 / num_k * np.ones((num_k))
if aTB.Atoms.spindeg:
    kps_wt *= 2
num_e = 1.2
num_band_max = 2

# GPARAMBANDS.h5
h1e_list = [np.array([[0,0,0.25,0],[0,0,0,0.25],[0.25,0,0,0],[0,0.25,0,0]], \
        dtype=np.complex)]
ginput.save_gparambands(kps_wt, num_e, num_band_max, h1e_list=h1e_list)
# GPARAM.h5
sigma = np.zeros((4,4), dtype=np.int32)
sigma[0::2, 0::2] = np.arange(1,5).reshape((2,2))
sigma[1::2, 1::2] = sigma[0::2, 0::2]
v2e = np.zeros((4, 4, 4, 4), dtype=np.complex)
v2e[0, 0, 0, 0] = v2e[0, 0, 1, 1] = v2e[1, 1, 0, 0] = v2e[1, 1, 1, 1] = 8.0
v2e[2:, 2:, 2:, 2:] = v2e[0:2, 0:2, 0:2, 0:2]
ryd2ev = s_const.physical_constants['Rydberg constant times hc in eV'][0]
sz = np.asarray(np.diag((1,-1,1,-1)),dtype=np.complex)

ginput.save_gparam(na2_list=[4], iembeddiag=-1, imix=-1, sigma_list=[sigma],
        v2e_list=[v2e], sz_list=[sz], nval_top_list=[4])

# BAREHAM_0.h5
aTB.save_bareham(kps)



# -*- coding: utf-8 -*-
from __future__ import print_function
from pygrisb.symm.unitary import comp_sph_harm_to_relativistic_harm

'''
Get the matrix representation of angular momentum operator
in complex spherical harmonics (CSH) basis.
The orbital index is faster and spin goes up first where applicable.
'''

import numpy as np
from scipy.linalg import block_diag
from pygrisb.math.matrix_util import trans_orbital_fast_to_spin_fast


def get_J_generator(l_list, iso):
    '''
    Get rotation generator J vector from a l_list with CSH basis.
    For iso=2, the fast index of the basis is orbital,
    and the spin goes up first then down.
    '''
    if iso == 1:
        # need L-vector for one spin-block only
        # because of no spin-orbit interaction.
        return get_L_vector(l_list)
    else:
        return get_J_vector(l_list)


def get_J_vector(l_list):
    '''get the total angular momentum vector in CSH basis,
    with orbital index faster and spin running through [up, down].
    '''
    Lx, Ly, Lz = get_L_vector(l_list, iso=2)
    Sx, Sy, Sz = get_S_vector(l_list)
    return [Lx + Sx, Ly + Sy, Lz + Sz]


def get_JU_relat_sph_harm_cg(l_list):
    '''get J vector in relativisitic spherical harmonics (RSH) basis,
    and the unitary transformation U from complex spherical harmonics
    (CSH) to RSH using clebsch-gordan coefficients.
    '''
    for i, _l in enumerate(l_list):
        _Jx, _Jy, _Jz =  get_J_vector(_l)
        v = comp_sph_harm_to_relativistic_harm((2*_l+1)*2)
        _Jx, _Jy, _Jz = v.T.conj().dot(_Jx).dot(v), \
                v.T.conj().dot(_Jy).dot(v), v.T.conj().dot(_Jz).dot(v)
        # check Jz
        for j in range(2*_l):
            jz = j-(_l-0.5)
            if abs(jz-_Jz[j,j]) > 1.e-6:
                raise ValueError("jz = {} vs expected {}!".format(\
                        _Jz[j], jz))
        for j in range(2*_l+2):
            jz = j-(_l+0.5)
            if abs(jz-_Jz[j+2*_l,j+2*_l]) > 1.e-6:
                raise ValueError("jz = {} vs expected {}!".format(\
                        _Jz[j+2*_l,j+2*_l], jz))
        if i == 0:
            Jx, Jy, Jz = _Jx, _Jy, _Jz
            u_trans = v
        else:
            Jx, Jy, Jz, u_trans = block_diag(Jx, _Jx), block_diag(Jy, _Jy), \
                    block_diag(Jz, _Jz), block_diag(u_trans, v)
    return [Jx, Jy, Jz], u_trans


def get_JU_relat_sph_harm_random_phase(l_list):
    '''get J vector in relativisitic spherical harmonics (RSH) basis,
    and the unitary transformation U from complex spherical harmonics
    (CSH) to RSH.
    There is a random phase associated RSH due to diagonalization.
    '''
    for i, _l in enumerate(l_list):
        _Jx, _Jy, _Jz =  get_J_vector(_l)
        Jsq = _Jx.dot(_Jx) + _Jy.dot(_Jy) + _Jz.dot(_Jz)
        w, v = np.linalg.eigh(Jsq)

        # from j(j+1) to j
        w = map(lambda x: np.sqrt(x+0.25)-0.5, w)

        if _l > 0:
            n1 = int(2*(_l - .5) + 1.1)
            if np.max(np.abs(w[:n1] - w[0])) > 1.e-6 or \
                    np.max(np.abs(w[n1:] - w[n1])) > 1.e-6 or \
                    np.abs(w[n1] - w[0] - 1) > 1.e-6:
                raise_j_list_error(w, 'j_list')

            # l-1/2 block
            Jz1 = v[:,:n1].T.conj().dot(_Jz).dot(v[:,:n1])
            w1, v1 = np.linalg.eigh(Jz1)
            if np.max(np.abs(w1[1:] - w1[:-1] - 1.)) > 1.e-6:
                raise_j_list_error(w1, 'jz1_list')
            v[:, :n1] = v[:, :n1].dot(v1)
        else:
            n1 = 0

        # l+1/2 block
        Jz1 = v[:,n1:].T.conj().dot(_Jz).dot(v[:,n1:])
        w1, v1 = np.linalg.eigh(Jz1)
        if np.max(np.abs(w1[1:] - w1[:-1] - 1.)) > 1.e-6:
            raise_j_list_error(w1, 'jz2_list')
        v[:, n1:] = v[:, n1:].dot(v1)
        _Jx, _Jy, _Jz = v.T.conj().dot(_Jx).dot(v), \
                v.T.conj().dot(_Jy).dot(v), v.T.conj().dot(_Jz).dot(v)
        if i == 0:
            Jx, Jy, Jz = _Jx, _Jy, _Jz
            u_trans = v
        else:
            Jx, Jy, Jz, u_trans = block_diag(Jx, _Jx), block_diag(Jy, _Jy), \
                    block_diag(Jz, _Jz), block_diag(u_trans, v)
    return [Jx, Jy, Jz], u_trans


def raise_j_list_error(j_list, head):
    msg = 'error in {}:\n'.format(head)
    for k, j in enumerate(j_list):
        msg += ' {:6.2f}'.format(j)
        if k % 5 == 4:
            msg += '/n'
    raise ValueError(msg)


def get_L_vector(l, iso=1):
    '''
    Get matrix representation of L-vector.
    '''
    if iso == 1:
        # for spin-up block
        try:
            for i, _l in enumerate(l):
                _Lx, _Ly, _Lz = get_L_vector(_l)
                if i == 0:
                    Lx, Ly, Lz = _Lx.copy(), _Ly.copy(), _Lz.copy()
                else:
                    Lx, Ly, Lz = block_diag(Lx, _Lx), block_diag(Ly, _Ly), \
                        block_diag(Lz, _Lz)
            return [Lx, Ly, Lz]
        # excepct single l
        except TypeError:
            Lz = get_matrix_Lz_CSH(l)
            Lp = get_matrix_Lp_CSH(l)
            _, Lx, Ly = get_other_op(Lz, Lp)
            return [Lx, Ly, Lz]
    else:
        # get spin-up block first
        Lx, Ly, Lz = get_L_vector(l)
        # add spin-down block
        Lx, Ly, Lz = block_diag(Lx, Lx), block_diag(Ly, Ly), \
                block_diag(Lz, Lz)
        return [Lx, Ly, Lz]


def get_Lp_coef(l, m):
    '''
    L^\dagger |l, m> = Lp_coef |l, m+1>.
    '''
    return np.sqrt(l * (l + 1) - m * (m + 1))


def get_diag_Lz_CSH(l):
    '''
    Get diagonal elements of Lz for Complex spherical Harmomnics basis.
    '''
    Lz = []
    for j in range(int(2 * l + 1)):
        Lz.append(-l + j)
    return Lz


def get_matrix_Lz_CSH(l):
    '''
    Get matrix_{z}.
    '''
    return np.diag(get_diag_Lz_CSH(l))


def get_matrix_Lp_CSH(l, m_rising=True):
    '''
    Get matrix L_{\dagger}.
    '''
    lm = int(2 * l + 1.5)
    Lp = np.zeros([lm, lm])
    if m_rising:
        for i in range(int(2 * l + 0.5)):
            Lp[i + 1, i] = get_Lp_coef(l, -l + i)
    else:
        for i in range(int(2 * l + 0.5)):
            Lp[i, i + 1] = get_Lp_coef(l, l - i - 1)
    return Lp


def get_matrix_Sz_CSH_orbital_fast(l_list):
    '''
    Get matrix Sz for Complex spherical Harmomnics basis
    with spin-up + spin-down block.
    '''
    num_lm = np.sum(2*np.asarray(l_list) + 1)

    # spin-up + spin_dn. Wien2k convention.
    Sz = np.diag([ 0.5 for i in range(num_lm)] + \
            [-0.5 for i in range(num_lm)])
    return Sz


def get_matrix_Sp_CSH_spin_fast(l_list):
    '''
    Get matrix Sp for Complex spherical Harmomnics basis with spin-fast
    (up, down) index.
    '''
    num_lm = np.sum(2 * np.asarray(l_list) + 1)
    Sp_sub = get_matrix_Lp_CSH(0.5, m_rising=False)
    Sp_list = [Sp_sub for i in range(num_lm)]
    from scipy.linalg import block_diag
    return block_diag(*Sp_list)


def get_S_vector(l_list):
    '''
    Get S-vector in CSH basis with fast orbital index.
    '''
    Sz = get_matrix_Sz_CSH_orbital_fast(l_list)
    Sp = get_matrix_Sp_CSH_spin_fast(l_list)
    Sp = trans_orbital_fast_to_spin_fast(Sp, lback=True)
    _, Sx, Sy = get_other_op(Sz, Sp)
    return [Sx, Sy, Sz]


def get_other_op(Lz, Lp):
    '''
    Get other related operators given L_z and L^\dagger.
    '''
    Ln = Lp.conj().T
    Lx = (Lp + Ln) / 2.
    Ly = (Lp - Ln) / (2.j)
    return Ln, Lx, Ly


def get_l_list_from_string(code):
    '''
    Gt l list from letters, e.g., "spd" -> [0, 1, 2].
    '''
    l_list = []
    for c in code:
        l = get_l_from_char(c)
        if l >= 0:
            l_list.append(l)
    return l_list


def get_l_from_char(c):
    return {'s': 0, 'p': 1, 'd': 2, 'f': 3}.get(c, -1)


def get_complex_to_real_sph_harm(l):
    '''get the unitary transformation from compex spherical harmonics
    (Condon–Shortley phase convention) to real harmonics
    (https://en.wikipedia.org/wiki/Spherical_harmonics,
    consistent with VASP).
    '''
    mdim = 2*l + 1
    c2r = np.zeros((mdim, mdim), dtype=np.complex)
    for m in range(mdim):
        m_ = m - l
        if m_ > 0:
            c2r[ m_+l, m] = (-1)**m_/np.sqrt(2.)
            c2r[-m_+l, m] = 1./np.sqrt(2.)
        elif m_ == 0:
            c2r[l, l] = 1.
        else:
            c2r[ m_+l, m] = 1.j/np.sqrt(2.)
            c2r[-m_+l, m] = -1.j*(-1)**m_/np.sqrt(2.)
    return c2r


def get_jvec_comp_sph_harm_to_rel_harm(j_csh, l_list):
    '''get the J vector in complex spherical harmonics representation
    to relativistic harmonics representation.
    the transformation matrix is also returned.
    '''
    lmbase = 0
    for il, l in enumerate(l_list):
        lm2 = (l*2+1)*2
        jx = j_csh[0][lmbase:lmbase+lm2, lmbase:lmbase+lm2]
        jy = j_csh[1][lmbase:lmbase+lm2, lmbase:lmbase+lm2]
        jz = j_csh[2][lmbase:lmbase+lm2, lmbase:lmbase+lm2]
        lmbase += lm2
        j2 = jx.dot(jx) + jy.dot(jy) + jz.dot(jz)
        w, v = np.linalg.eigh(j2)
        jx = v.conj().T.dot(jx).dot(v)
        jy = v.conj().T.dot(jy).dot(v)
        jz = v.conj().T.dot(jz).dot(v)
        if il == 0:
            j_rel = [jx, jy, jz]
            u_csh2rel = v
        else:
            j_rel[0] = block_diag(j_rel[0], jx)
            j_rel[1] = block_diag(j_rel[1], jy)
            j_rel[2] = block_diag(j_rel[2], jz)
            u_csh2rel = block_diag(u_csh2rel, v)
    return j_rel, u_csh2rel



if __name__ == "__main__":
    Lx, Ly, Lz = get_L_vector(0)
    print(Lx, Ly, Lz)



import numpy as np
from scipy.linalg import block_diag
from pygrisb.math.matrix_util import trans_orbital_fast_to_spin_fast
from pygrisb.symm.angular_momentum_1p import get_J_generator, \
        get_JU_relat_sph_harm_cg
import pygrisb.symm.atom_symm as atsym


def get_self_energy(l_list, ispin, orbital_pol, cf, iso, rotations=None):
    '''
    Dispatcher.
    '''
    if 'y' in orbital_pol and iso == 2:
        return get_self_energy_op_soc(l_list)
    elif 'y' in cf and iso == 2:
        return get_self_energy_cf_soc(l_list, rotations)
    elif iso == 2:
        return get_self_energy_soc(l_list, ispin)
    elif 'y' in orbital_pol:
        return get_self_energy_op_nosoc(l_list, ispin)
    elif 'y' in cf:
        return get_self_energy_cf_nosoc(l_list, ispin, rotations)
    else:
        return get_self_energy_average(l_list, ispin)


def get_self_energy_op_soc(l_list):
    '''
    Get the self energy structure in the case of fully orbital symmetry
    breaking and with spin-orbit interaction.
    '''
    dim_tot = np.sum(2 * (2 * np.array(l_list) + 1))
    self_energy = np.arange(dim_tot * dim_tot, dtype=int) + 1
    self_energy = self_energy.reshape((dim_tot, dim_tot))
    U = np.identity(dim_tot, dtype=complex)
    return None, U, self_energy


def get_self_energy_op_nosoc(l_list, ispin):
    '''
    Get the self energy structure in the case of fully orbital symmetry
    breaking but with negligible spin-orbit interaction.
    '''
    dim_t = int(np.sum(2 * np.array(l_list) + 1) + 0.5)
    m_half = np.arange(dim_t * dim_t, dtype=int) + 1
    m_half = m_half.reshape((dim_t, dim_t))
    if ispin == 2:
        shift = np.max(m_half)
    else:
        shift = 0
    from scipy.linalg import block_diag
    self_energy = block_diag(m_half, m_half + shift)

    # spin-fast-index convention
    self_energy = trans_orbital_fast_to_spin_fast(self_energy)
    Uhalf = np.identity(dim_t)

    # {{orbs}_up, {orbs}_dn} -> {{up,dn}_{orbs}}
    U = np.zeros((dim_t*2, dim_t*2), dtype=complex)
    U[:dim_t,0::2] = Uhalf
    U[dim_t:,1::2] = Uhalf
    return None, U, self_energy


def get_self_energy_cf_soc(l_list, rotations):
    '''
    Get the self energy structure in the case of crystal field splitting
    and with spin-orbit interaction.
    '''
    j_rel, u_csh2rel = get_JU_relat_sph_harm_cg(l_list)
    J, U, self_energy = atsym.get_atom_Jnew(rotations, j_rel)
    U = u_csh2rel.dot(U)
    return J, U, self_energy


def get_self_energy_cf_nosoc(l_list, ispin, rotations):
    '''
    Get the self energy structure in the case of crystal field splitting
    and without spin-orbit interaction.
    '''
    Jorig = get_J_generator(l_list, iso=1)
    Jhalf, Uhalf, self_energy_half = atsym.get_atom_Jnew(rotations, Jorig)

    if ispin == 2:
        shift = np.max(self_energy_half)
    else:
        shift = 0

    self_energy = block_diag(
        self_energy_half, shift_self_energy(self_energy_half, shift))

    # spin-fast-index convfention
    self_energy = trans_orbital_fast_to_spin_fast(self_energy)

    dim_t = Uhalf.shape[0]
    J = []
    for Jh1 in Jhalf:
        J1 = np.zeros(self_energy.shape, dtype=np.complex)
        J1[0::2,0::2] = Jh1
        J1[1::2,1::2] = Jh1
        J.append(J1)

    U = np.zeros(self_energy.shape, dtype=np.complex)
    U[:dim_t,0::2] = Uhalf
    U[dim_t:,1::2] = Uhalf
    return J, U, self_energy


def get_self_energy_average(l_list, ispin):
    '''
    Get the self energy structure in the case of negligible crystal field
    splitting and without spin-orbit interaction.
    '''
    diag_elem = []
    elem_base = 1
    for l in l_list:
        diag_elem += [elem_base for i in range(2 * l + 1)]
        elem_base += 1
    elem_base -= 1
    if ispin == 2:
        diag_elem += [i + elem_base for i in diag_elem]
    else:
        diag_elem += diag_elem
    self_energy = np.diag(diag_elem)
    self_energy = trans_orbital_fast_to_spin_fast(self_energy)
    U = np.identity(len(self_energy), dtype=complex)
    n_half = U.shape[0]/2
    utrans = np.zeros_like(U)

    # Similarly transform to spin-fast convention.
    utrans[:,::2] = U[:,:n_half]
    utrans[:,1::2] = utrans[:,::2]
    return None, utrans, self_energy


def get_self_energy_soc(l_list, ispin):
    '''
    Get the self energy structure in the case that spin-orbit interaction
    is dominant.
    '''
    diag_elem = []
    elem_base = 1
    for l in l_list:
        if l == 0:
            if ispin == 1:
                elem = [elem_base, elem_base]
            else:
                elem = [elem_base, elem_base + 1]
        else:
            elem = []
            if ispin == 1:
                elem += [elem_base for i in \
                        range(int(2 * (l - 0.49)) + 1)]
                elem_base = max(elem) + 1
                elem += [elem_base for i in \
                        range(int(2 * (l + 0.51)) + 1)]
            else:
                elem += [elem_base + i for i in \
                        range(int(2 * (l - 0.49)) + 1)]
                elem_base = max(elem) + 1
                elem += [elem_base + i for i in \
                        range(int(2 * (l + 0.51)) + 1)]
        diag_elem += elem
        elem_base = max(elem) + 1
    self_energy = np.diag(diag_elem)

    J, U = get_JU_relat_sph_harm_cg(l_list)
    return J, U, self_energy


def shift_self_energy(self_energy, shift):
    '''
    Shift non-zero self-energy elements by shift.
    '''
    res = np.zeros_like(self_energy)
    spots = np.where(self_energy > 0.1)
    res[spots] = self_energy[spots] + shift
    return res



if __name__ == "__main__":
    '''
    A test.
    '''
    l_list = [1, 1]
    ispin = 1
    orbital_pol = 'n'
    cf = 'y'
    iso = 1
    rotations = [[[1, 0, 0], [0, 1, 0], [0, 0, 1]]]
    U, self_energy = get_self_energy(
        l_list, ispin, orbital_pol, cf, iso, rotations)
    print(" self_energy = ")
    print(self_energy)

    # H cluster
    symbols = ['H', 'H', 'H', 'H', 'H', 'H']
    positions = [[0., 0., 0.], [2.0, 0., 0.], [0., 2.0, 0.0],
                 [0., -2.0, 0.], [0., 0., 2.0], [0., 0., -2.0]]
    from pygrisb.gutz.molecule import xyz_get_rot_list
    rotations = xyz_get_rot_list(symbols, positions, log='screen')
    print("H-dimer rotations:")
    print(rotations)
    l_list = [0, 0, 1]
    orbital_pol = 'n'
    cf = 'y'
    J, U, self_energy = get_self_energy(
        l_list, ispin, orbital_pol, cf, iso, rotations)
    print(" self_energy = ")
    print(self_energy)
